library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity buffered_line_generator_tester is
end buffered_line_generator_tester;

architecture test of buffered_line_generator_tester is
	signal clock : std_logic := '0';
	signal reset : std_logic := '0';
	
	signal row : signed (8 downto 0) := "011001101";
	
	signal rect_memory_address : unsigned (7 downto 0) := (others => '0');
	signal rect_memory_address_delayed : unsigned (7 downto 0) := (others => '0');
	signal rect_memory_data : unsigned (7 downto 0) := (others => '0');
	
	-- Access to line buffer
	signal screen_col_to_read : unsigned(8 downto 0) := (others => '0');
	signal screen_col_color : std_logic_vector (17 downto 0) := (others => '0');
	
	-- unit testing signals
	signal done : std_logic := '0';
begin
	process
	begin
		loop
			wait for 10 ns;
			clock <= '1';
			wait for 10 ns;
			clock <= '0';
		end loop;
	end process;
	
	process (clock)
	begin
		rect_memory_address_delayed <= rect_memory_address;
	end process;
	
	process (rect_memory_address)
	begin
		if rect_memory_address_delayed = "00000000" then
			rect_memory_data <= "00000000";
		elsif rect_memory_address_delayed = "00000001" then
			rect_memory_data <= "00000000";
		elsif rect_memory_address_delayed = "00000010" then
			rect_memory_data <= "00000000";
		elsif rect_memory_address_delayed = "00000011" then
			rect_memory_data <= "11111110";
		elsif rect_memory_address_delayed = "00000100" then
			rect_memory_data <= "11111110";
		elsif rect_memory_address_delayed = "00000101" then
			rect_memory_data <= "00000000";
		elsif rect_memory_address_delayed = "00000110" then
			rect_memory_data <= "00111100";
		elsif rect_memory_address_delayed = "00000111" then
			rect_memory_data <= "11111110";
		elsif rect_memory_address_delayed = "00001000" then
			rect_memory_data <= "00000000";
		elsif rect_memory_address_delayed = "00001001" then
			rect_memory_data <= "00000000";
		elsif rect_memory_address_delayed = "00001010" then
			rect_memory_data <= "11111110";
		elsif rect_memory_address_delayed = "00001011" then
			rect_memory_data <= "11111110";
		elsif rect_memory_address_delayed = "00001100" then
			rect_memory_data <= "11111110";
		elsif rect_memory_address_delayed = "00001101" then
			rect_memory_data <= "00000011";
		elsif rect_memory_address_delayed = "00001110" then
			rect_memory_data <= "11110001";
		elsif rect_memory_address_delayed = "00001111" then
			rect_memory_data <= "10100001";
		elsif rect_memory_address_delayed = "00010000" then
			rect_memory_data <= "00000001";
		elsif rect_memory_address_delayed = "00010001" then
			rect_memory_data <= "11111011";
		elsif rect_memory_address_delayed = "00010010" then
			rect_memory_data <= "10110101";
		elsif rect_memory_address_delayed = "00010011" then
			rect_memory_data <= "00000001";
		elsif rect_memory_address_delayed = "00010100" then
			rect_memory_data <= "00001100";
		elsif rect_memory_address_delayed = "00010101" then
			rect_memory_data <= "11110001";
		elsif rect_memory_address_delayed = "00010110" then
			rect_memory_data <= "10100001";
		elsif rect_memory_address_delayed = "00010111" then
			rect_memory_data <= "00001011";
		elsif rect_memory_address_delayed = "00011000" then
			rect_memory_data <= "11111011";
		elsif rect_memory_address_delayed = "00011001" then
			rect_memory_data <= "10110101";
		elsif rect_memory_address_delayed = "00011010" then
			rect_memory_data <= "00001011";
		elsif rect_memory_address_delayed = "00011011" then
			rect_memory_data <= "00001100";
		elsif rect_memory_address_delayed = "00011100" then
			rect_memory_data <= "11110001";
		elsif rect_memory_address_delayed = "00011101" then
			rect_memory_data <= "10100001";
		elsif rect_memory_address_delayed = "00011110" then
			rect_memory_data <= "00010101";
		elsif rect_memory_address_delayed = "00011111" then
			rect_memory_data <= "11111011";
		elsif rect_memory_address_delayed = "00100000" then
			rect_memory_data <= "10110101";
		elsif rect_memory_address_delayed = "00100001" then
			rect_memory_data <= "00010101";
		elsif rect_memory_address_delayed = "00100010" then
			rect_memory_data <= "00001100";
		elsif rect_memory_address_delayed = "00100011" then
			rect_memory_data <= "11001000";
		elsif rect_memory_address_delayed = "00100100" then
			rect_memory_data <= "00110010";
		elsif rect_memory_address_delayed = "00100101" then
			rect_memory_data <= "11111110";
		elsif rect_memory_address_delayed = "00100110" then
			rect_memory_data <= "11001000";
		elsif rect_memory_address_delayed = "00100111" then
			rect_memory_data <= "11001000";
		elsif rect_memory_address_delayed = "00101000" then
			rect_memory_data <= "11111011";
		elsif rect_memory_address_delayed = "00101001" then
			rect_memory_data <= "00110011";
		else
			rect_memory_data <= x"00";
		end if;
	end process;
	
	generator: entity work.buffered_line_generator
    port map (
		clock => clock,
		row => row,
		camera_x => "011100111", camera_y => "010101011", camera_z => "000001010",
		rect_memory_address => rect_memory_address,
		rect_memory_data => rect_memory_data,
		screen_col => screen_col_to_read,
		screen_col_color => screen_col_color,
		done => done
	);
	
	process
	begin
		row <= "010111101";
		wait until done = '1'; row <= "010111100";
		wait until done = '1'; row <= "010111011";
		wait until done = '1'; row <= "010111010";
		wait until done = '1'; row <= "010111001";
		wait until done = '1'; row <= "010111000";
		wait until done = '1'; row <= "010110111";
		wait until done = '1'; row <= "010110110";
		wait until done = '1'; row <= "010110101";
		wait until done = '1'; row <= "010110100";
		wait until done = '1'; row <= "010110011";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -159 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -158 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -157 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -156 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -155 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -154 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -153 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -152 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -151 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -150 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -149 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -148 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -147 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -146 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -145 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -144 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -143 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -142 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -141 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -140 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -139 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -138 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -137 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -136 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -135 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -134 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -133 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -132 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -131 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -130 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -129 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -128 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -127 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -126 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -125 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -124 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -123 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -122 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -121 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -120 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -119 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -118 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -117 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -116 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -115 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -114 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -113 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -112 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -111 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -110 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -109 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -108 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -107 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -106 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -105 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -104 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -103 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -102 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -101 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -100 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -99 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -98 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -97 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -96 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -95 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -94 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -93 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -92 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -91 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -90 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -89 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -88 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -87 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -86 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -85 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -84 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -83 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -82 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -81 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -80 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -79 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -78 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -77 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -76 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -75 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -74 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -73 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -72 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -71 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -70 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -69 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -68 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -67 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -66 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -65 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -64 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -63 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -62 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -61 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -60 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -59 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -58 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -57 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -56 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -55 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -54 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -53 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -52 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -51 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -50 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -49 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -48 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -47 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -46 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -45 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -44 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -43 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -42 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -41 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -40 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -39 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -38 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -37 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -36 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -35 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -34 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -33 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -32 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -31 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -30 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -29 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -28 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -27 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -26 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -25 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -24 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -23 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -22 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -21 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -20 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -19 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -18 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -17 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -16 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -15 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -14 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -13 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -12 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -11 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -10 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -9 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -8 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -7 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -6 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -5 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -4 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -3 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -2 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col -1 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 0 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 1 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 2 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 3 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 4 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 5 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 6 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 7 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 8 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 9 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 10 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 11 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 12 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 13 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 14 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 15 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 16 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 17 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 18 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 19 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 20 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 21 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 22 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 23 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 24 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 25 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 26 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 27 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 28 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 29 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 30 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 31 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 32 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 33 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 34 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 35 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 36 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 37 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 38 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 39 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 40 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 41 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 42 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 43 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 44 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 45 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 46 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 47 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 48 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 49 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 50 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 51 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 52 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 53 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 54 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 55 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 56 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 57 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 58 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 59 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 60 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 61 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 62 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 63 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 64 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 65 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 66 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 67 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 68 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 69 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 70 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 71 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 72 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 73 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 74 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 75 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 76 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 77 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 78 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 79 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 80 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 81 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 82 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 83 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 84 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 85 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 86 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 87 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 88 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 89 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 90 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 91 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 92 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 93 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 94 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 95 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 96 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 97 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 98 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 99 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 100 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 101 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 102 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 103 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 104 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 105 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 106 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 107 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 108 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 109 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 110 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 111 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 112 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 113 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 114 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 115 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 116 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 117 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 118 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 119 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 120 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 121 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 122 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 123 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 124 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 125 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 126 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 127 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 128 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 129 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 130 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 131 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 132 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 133 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 134 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 135 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 136 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 137 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 138 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 139 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 140 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 141 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 142 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 143 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 144 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 145 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 146 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 147 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 148 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 149 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 150 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 151 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 152 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 153 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 154 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 155 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 156 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 157 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 158 r,g,b = 0, 0, 46" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101110" report "row 180, col 159 r,g,b = 0, 0, 46" severity error;
		wait until done = '1'; row <= "010110010";
		wait until done = '1'; row <= "010110001";
		wait until done = '1'; row <= "010110000";
		wait until done = '1'; row <= "010101111";
		wait until done = '1'; row <= "010101110";
		wait until done = '1'; row <= "010101101";
		wait until done = '1'; row <= "010101100";
		wait until done = '1'; row <= "010101011";
		wait until done = '1'; row <= "010101010";
		wait until done = '1'; row <= "010101001";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -159 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -158 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -157 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -156 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -155 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -154 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -153 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -152 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -151 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -150 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -149 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -148 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -147 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -146 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -145 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -144 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -143 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -142 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -141 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -140 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -139 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -138 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -137 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -136 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -135 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -134 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -133 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -132 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -131 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -130 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -129 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -128 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -127 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -126 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -125 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -124 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -123 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -122 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -121 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -120 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -119 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -118 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -117 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -116 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -115 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -114 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -113 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -112 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -111 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -110 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -109 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -108 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -107 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -106 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -105 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -104 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -103 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -102 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -101 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -100 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -99 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -98 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -97 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -96 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -95 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -94 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -93 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -92 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -91 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -90 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -89 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -88 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -87 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -86 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -85 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -84 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -83 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -82 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -81 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -80 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -79 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -78 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -77 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -76 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -75 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -74 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -73 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -72 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -71 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -70 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -69 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -68 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -67 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -66 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -65 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -64 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -63 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -62 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -61 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -60 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -59 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -58 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -57 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -56 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -55 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -54 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -53 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -52 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -51 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -50 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -49 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -48 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -47 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -46 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -45 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -44 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -43 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -42 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -41 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -40 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -39 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -38 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -37 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -36 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -35 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -34 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -33 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -32 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -31 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -30 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -29 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -28 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -27 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -26 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -25 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -24 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -23 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -22 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -21 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -20 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -19 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -18 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -17 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -16 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -15 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -14 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -13 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -12 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -11 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -10 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -9 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -8 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -7 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -6 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -5 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -4 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -3 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -2 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col -1 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 0 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 1 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 2 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 3 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 4 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 5 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 6 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 7 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 8 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 9 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 10 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 11 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 12 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 13 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 14 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 15 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 16 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 17 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 18 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 19 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 20 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 21 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 22 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 23 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 24 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 25 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 26 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 27 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 28 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 29 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 30 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 31 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 32 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 33 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 34 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 35 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 36 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 37 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 38 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 39 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 40 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 41 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 42 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 43 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 44 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 45 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 46 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 47 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 48 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 49 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 50 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 51 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 52 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 53 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 54 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 55 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 56 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 57 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 58 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 59 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 60 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 61 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 62 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 63 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 64 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 65 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 66 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 67 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 68 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 69 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 70 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 71 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 72 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 73 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 74 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 75 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 76 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 77 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 78 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 79 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 80 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 81 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 82 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 83 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 84 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 85 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 86 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 87 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 88 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 89 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 90 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 91 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 92 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 93 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 94 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 95 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 96 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 97 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 98 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 99 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 100 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 101 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 102 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 103 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 104 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 105 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 106 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 107 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 108 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 109 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 110 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 111 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 112 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 113 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 114 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 115 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 116 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 117 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 118 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 119 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 120 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 121 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 122 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 123 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 124 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 125 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 126 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 127 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 128 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 129 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 130 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 131 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 132 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 133 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 134 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 135 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 136 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 137 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 138 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 139 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 140 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 141 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 142 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 143 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 144 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 145 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 146 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 147 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 148 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 149 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 150 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 151 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 152 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 153 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 154 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 155 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 156 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 157 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 158 r,g,b = 0, 0, 45" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101101" report "row 170, col 159 r,g,b = 0, 0, 45" severity error;
		wait until done = '1'; row <= "010101000";
		wait until done = '1'; row <= "010100111";
		wait until done = '1'; row <= "010100110";
		wait until done = '1'; row <= "010100101";
		wait until done = '1'; row <= "010100100";
		wait until done = '1'; row <= "010100011";
		wait until done = '1'; row <= "010100010";
		wait until done = '1'; row <= "010100001";
		wait until done = '1'; row <= "010100000";
		wait until done = '1'; row <= "010011111";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -159 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -158 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -157 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -156 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -155 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -154 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -153 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -152 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -151 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -150 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -149 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -148 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -147 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -146 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -145 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -144 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -143 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -142 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -141 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -140 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -139 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -138 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -137 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -136 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -135 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -134 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -133 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -132 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -131 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -130 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -129 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -128 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -127 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -126 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -125 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -124 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -123 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -122 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -121 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -120 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -119 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -118 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -117 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -116 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -115 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -114 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -113 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -112 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -111 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -110 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -109 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -108 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -107 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -106 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -105 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -104 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -103 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -102 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -101 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -100 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -99 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -98 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -97 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -96 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -95 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -94 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -93 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -92 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -91 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -90 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -89 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -88 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -87 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -86 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -85 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -84 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -83 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -82 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -81 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -80 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -79 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -78 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -77 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -76 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -75 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -74 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -73 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -72 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -71 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -70 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -69 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -68 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -67 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -66 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -65 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -64 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -63 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -62 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -61 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -60 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -59 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -58 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -57 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -56 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -55 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -54 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -53 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -52 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -51 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -50 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -49 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -48 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -47 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -46 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -45 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -44 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -43 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -42 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -41 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -40 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -39 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -38 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -37 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -36 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -35 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -34 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -33 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -32 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -31 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -30 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -29 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -28 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -27 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -26 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -25 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -24 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -23 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -22 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -21 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -20 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -19 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -18 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -17 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -16 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -15 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -14 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -13 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -12 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -11 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -10 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -9 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -8 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -7 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -6 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -5 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -4 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -3 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -2 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col -1 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 0 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 1 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 2 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 3 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 4 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 5 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 6 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 7 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 8 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 9 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 10 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 11 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 12 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 13 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 14 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 15 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 16 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 17 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 18 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 19 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 20 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 21 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 22 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 23 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 24 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 25 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 26 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 27 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 28 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 29 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 30 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 31 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 32 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 33 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 34 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 35 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 36 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 37 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 38 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 39 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 40 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 41 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 42 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 43 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 44 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 45 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 46 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 47 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 48 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 49 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 50 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 51 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 52 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 53 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 54 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 55 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 56 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 57 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 58 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 59 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 60 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 61 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 62 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 63 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 64 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 65 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 66 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 67 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 68 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 69 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 70 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 71 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 72 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 73 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 74 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 75 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 76 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 77 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 78 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 79 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 80 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 81 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 82 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 83 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 84 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 85 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 86 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 87 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 88 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 89 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 90 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 91 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 92 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 93 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 94 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 95 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 96 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 97 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 98 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 99 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 100 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 101 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 102 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 103 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 104 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 105 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 106 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 107 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 108 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 109 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 110 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 111 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 112 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 113 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 114 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 115 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 116 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 117 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 118 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 119 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 120 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 121 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 122 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 123 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 124 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 125 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 126 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 127 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 128 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 129 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 130 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 131 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 132 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 133 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 134 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 135 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 136 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 137 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 138 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 139 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 140 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 141 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 142 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 143 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 144 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 145 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 146 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 147 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 148 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 149 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 150 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 151 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 152 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 153 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 154 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 155 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 156 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 157 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 158 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 160, col 159 r,g,b = 0, 0, 44" severity error;
		wait until done = '1'; row <= "010011110";
		wait until done = '1'; row <= "010011101";
		wait until done = '1'; row <= "010011100";
		wait until done = '1'; row <= "010011011";
		wait until done = '1'; row <= "010011010";
		wait until done = '1'; row <= "010011001";
		wait until done = '1'; row <= "010011000";
		wait until done = '1'; row <= "010010111";
		wait until done = '1'; row <= "010010110";
		wait until done = '1'; row <= "010010101";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -159 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -158 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -157 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -156 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -155 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -154 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -153 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -152 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -151 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -150 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -149 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -148 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -147 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -146 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -145 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -144 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -143 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -142 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -141 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -140 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -139 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -138 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -137 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -136 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -135 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -134 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -133 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -132 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -131 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -130 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -129 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -128 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -127 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -126 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -125 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -124 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -123 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -122 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -121 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -120 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -119 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -118 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -117 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -116 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -115 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -114 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -113 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -112 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -111 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -110 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -109 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -108 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -107 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -106 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -105 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -104 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -103 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -102 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -101 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -100 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -99 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -98 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -97 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -96 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -95 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -94 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -93 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -92 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -91 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -90 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -89 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -88 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -87 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -86 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -85 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -84 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -83 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -82 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -81 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -80 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -79 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -78 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -77 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -76 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -75 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -74 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -73 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -72 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -71 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -70 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -69 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -68 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -67 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -66 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -65 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -64 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -63 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -62 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -61 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -60 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -59 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -58 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -57 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -56 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -55 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -54 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -53 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -52 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -51 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -50 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -49 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -48 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -47 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -46 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -45 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -44 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -43 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -42 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -41 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -40 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -39 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -38 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -37 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -36 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -35 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -34 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -33 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -32 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -31 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -30 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -29 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -28 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -27 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -26 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -25 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -24 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -23 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -22 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -21 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -20 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -19 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -18 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -17 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -16 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -15 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -14 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -13 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -12 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -11 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -10 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -9 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -8 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -7 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -6 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -5 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -4 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -3 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -2 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col -1 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 0 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 1 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 2 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 3 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 4 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 5 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 6 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 7 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 8 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 9 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 10 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 11 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 12 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 13 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 14 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 15 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 16 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 17 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 18 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 19 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 20 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 21 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 22 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 23 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 24 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 25 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 26 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 27 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 28 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 29 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 30 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 31 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 32 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 33 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 34 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 35 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 36 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 37 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 38 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 39 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 40 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 41 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 42 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 43 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 44 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 45 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 46 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 47 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 48 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 49 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 50 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 51 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 52 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 53 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 54 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 55 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 56 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 57 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 58 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 59 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 60 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 61 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 62 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 63 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 64 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 65 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 66 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 67 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 68 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 69 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 70 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 71 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 72 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 73 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 74 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 75 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 76 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 77 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 78 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 79 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 80 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 81 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 82 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 83 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 84 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 85 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 86 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 87 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 88 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 89 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 90 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 91 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 92 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 93 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 94 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 95 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 96 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 97 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 98 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 99 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 100 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 101 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 102 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 103 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 104 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 105 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 106 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 107 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 108 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 109 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 110 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 111 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 112 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 113 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 114 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 115 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 116 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 117 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 118 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 119 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 120 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 121 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 122 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 123 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 124 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 125 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 126 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 127 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 128 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 129 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 130 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 131 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 132 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 133 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 134 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 135 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 136 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 137 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 138 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 139 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 140 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 141 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 142 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 143 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 144 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 145 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 146 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 147 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 148 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 149 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 150 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 151 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 152 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 153 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 154 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 155 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 156 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 157 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 158 r,g,b = 0, 0, 44" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101100" report "row 150, col 159 r,g,b = 0, 0, 44" severity error;
		wait until done = '1'; row <= "010010100";
		wait until done = '1'; row <= "010010011";
		wait until done = '1'; row <= "010010010";
		wait until done = '1'; row <= "010010001";
		wait until done = '1'; row <= "010010000";
		wait until done = '1'; row <= "010001111";
		wait until done = '1'; row <= "010001110";
		wait until done = '1'; row <= "010001101";
		wait until done = '1'; row <= "010001100";
		wait until done = '1'; row <= "010001011";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -159 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -158 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -157 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -156 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -155 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -154 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -153 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -152 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -151 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -150 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -149 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -148 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -147 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -146 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -145 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -144 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -143 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -142 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -141 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -140 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -139 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -138 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -137 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -136 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -135 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -134 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -133 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -132 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -131 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -130 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -129 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -128 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -127 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -126 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -125 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -124 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -123 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -122 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -121 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -120 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -119 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -118 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -117 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -116 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -115 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -114 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -113 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -112 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -111 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -110 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -109 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -108 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -107 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -106 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -105 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -104 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -103 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -102 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -101 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -100 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -99 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -98 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -97 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -96 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -95 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -94 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -93 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -92 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -91 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -90 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -89 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -88 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -87 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -86 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -85 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -84 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -83 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -82 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -81 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -80 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -79 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -78 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -77 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -76 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -75 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -74 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -73 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -72 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -71 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -70 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -69 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -68 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -67 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -66 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -65 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -64 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -63 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -62 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -61 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -60 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -59 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -58 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -57 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -56 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -55 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -54 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -53 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -52 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -51 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -50 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -49 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -48 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -47 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -46 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -45 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -44 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -43 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -42 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -41 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -40 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -39 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -38 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -37 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -36 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -35 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -34 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -33 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -32 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -31 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -30 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -29 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -28 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -27 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -26 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -25 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -24 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -23 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -22 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -21 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -20 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -19 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -18 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -17 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -16 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -15 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -14 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -13 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -12 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -11 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -10 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -9 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -8 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -7 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -6 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -5 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -4 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -3 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -2 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col -1 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 0 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 1 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 2 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 3 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 4 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 5 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 6 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 7 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 8 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 9 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 10 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 11 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 12 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 13 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 14 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 15 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 16 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 17 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 18 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 19 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 20 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 21 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 22 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 23 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 24 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 25 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 26 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 27 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 28 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 29 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 30 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 31 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 32 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 33 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 34 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 35 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 36 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 37 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 38 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 39 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 40 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 41 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 42 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 43 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 44 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 45 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 46 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 47 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 48 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 49 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 50 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 51 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 52 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 53 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 54 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 55 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 56 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 57 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 58 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 59 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 60 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 61 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 62 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 63 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 64 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 65 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 66 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 67 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 68 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 69 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 70 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 71 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 72 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 73 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 74 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 75 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 76 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 77 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 78 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 79 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 80 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 81 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 82 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 83 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 84 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 85 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 86 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 87 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 88 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 89 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 90 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 91 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 92 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 93 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 94 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 95 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 96 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 97 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 98 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 99 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 100 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 101 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 102 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 103 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 104 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 105 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 106 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 107 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 108 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 109 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 110 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 111 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 112 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 113 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 114 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 115 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 116 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 117 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 118 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 119 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 120 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 121 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 122 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 123 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 124 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 125 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 126 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 127 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 128 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 129 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 130 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 131 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 132 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 133 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 134 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 135 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 136 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 137 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 138 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 139 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 140 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 141 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 142 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 143 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 144 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 145 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 146 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 147 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 148 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 149 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 150 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 151 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 152 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 153 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 154 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 155 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 156 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 157 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 158 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 140, col 159 r,g,b = 0, 0, 43" severity error;
		wait until done = '1'; row <= "010001010";
		wait until done = '1'; row <= "010001001";
		wait until done = '1'; row <= "010001000";
		wait until done = '1'; row <= "010000111";
		wait until done = '1'; row <= "010000110";
		wait until done = '1'; row <= "010000101";
		wait until done = '1'; row <= "010000100";
		wait until done = '1'; row <= "010000011";
		wait until done = '1'; row <= "010000010";
		wait until done = '1'; row <= "010000001";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -159 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -158 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -157 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -156 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -155 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -154 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -153 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -152 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -151 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -150 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -149 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -148 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -147 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -146 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -145 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -144 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -143 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -142 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -141 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -140 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -139 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -138 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -137 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -136 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -135 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -134 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -133 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -132 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -131 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -130 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -129 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -128 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -127 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -126 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -125 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -124 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -123 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -122 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -121 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -120 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -119 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -118 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -117 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -116 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -115 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -114 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -113 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -112 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -111 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -110 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -109 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -108 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -107 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -106 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -105 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -104 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -103 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -102 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -101 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -100 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -99 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -98 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -97 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -96 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -95 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -94 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -93 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -92 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -91 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -90 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -89 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -88 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -87 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -86 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -85 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -84 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -83 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -82 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -81 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -80 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -79 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -78 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -77 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -76 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -75 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -74 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -73 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -72 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -71 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -70 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -69 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -68 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -67 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -66 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -65 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -64 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -63 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -62 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -61 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -60 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -59 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -58 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -57 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -56 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -55 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -54 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -53 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -52 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -51 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -50 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -49 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -48 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -47 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -46 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -45 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -44 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -43 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -42 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -41 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -40 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -39 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -38 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -37 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -36 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -35 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -34 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -33 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -32 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -31 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -30 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -29 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -28 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -27 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -26 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -25 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -24 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -23 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -22 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -21 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -20 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -19 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -18 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -17 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -16 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -15 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -14 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -13 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -12 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -11 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -10 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -9 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -8 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -7 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -6 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -5 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -4 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -3 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -2 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col -1 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 0 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 1 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 2 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 3 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 4 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 5 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 6 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 7 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 8 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 9 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 10 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 11 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 12 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 13 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 14 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 15 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 16 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 17 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 18 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 19 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 20 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 21 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 22 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 23 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 24 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 25 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 26 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 27 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 28 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 29 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 30 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 31 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 32 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 33 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 34 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 35 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 36 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 37 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 38 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 39 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 40 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 41 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 42 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 43 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 44 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 45 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 46 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 47 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 48 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 49 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 50 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 51 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 52 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 53 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 54 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 55 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 56 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 57 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 58 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 59 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 60 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 61 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 62 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 63 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 64 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 65 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 66 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 67 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 68 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 69 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 70 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 71 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 72 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 73 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 74 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 75 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 76 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 77 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 78 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 79 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 80 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 81 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 82 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 83 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 84 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 85 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 86 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 87 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 88 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 89 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 90 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 91 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 92 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 93 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 94 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 95 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 96 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 97 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 98 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 99 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 100 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 101 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 102 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 103 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 104 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 105 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 106 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 107 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 108 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 109 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 110 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 111 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 112 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 113 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 114 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 115 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 116 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 117 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 118 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 119 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 120 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 121 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 122 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 123 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 124 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 125 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 126 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 127 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 128 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 129 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 130 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 131 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 132 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 133 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 134 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 135 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 136 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 137 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 138 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 139 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 140 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 141 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 142 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 143 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 144 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 145 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 146 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 147 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 148 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 149 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 150 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 151 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 152 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 153 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 154 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 155 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 156 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 157 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 158 r,g,b = 0, 0, 43" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101011" report "row 130, col 159 r,g,b = 0, 0, 43" severity error;
		wait until done = '1'; row <= "010000000";
		wait until done = '1'; row <= "001111111";
		wait until done = '1'; row <= "001111110";
		wait until done = '1'; row <= "001111101";
		wait until done = '1'; row <= "001111100";
		wait until done = '1'; row <= "001111011";
		wait until done = '1'; row <= "001111010";
		wait until done = '1'; row <= "001111001";
		wait until done = '1'; row <= "001111000";
		wait until done = '1'; row <= "001110111";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -159 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -158 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -157 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -156 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -155 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -154 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -153 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -152 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -151 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -150 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -149 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -148 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -147 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -146 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -145 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -144 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -143 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -142 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -141 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -140 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -139 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -138 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -137 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -136 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -135 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -134 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -133 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -132 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -131 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -130 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -129 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -128 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -127 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -126 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -125 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -124 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -123 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -122 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -121 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -120 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -119 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -118 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -117 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -116 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -115 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -114 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -113 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -112 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -111 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -110 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -109 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -108 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -107 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -106 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -105 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -104 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -103 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -102 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -101 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -100 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -99 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -98 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -97 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -96 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -95 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -94 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -93 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -92 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -91 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -90 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -89 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -88 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -87 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -86 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -85 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -84 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -83 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -82 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -81 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -80 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -79 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -78 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -77 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -76 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -75 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -74 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -73 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -72 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -71 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -70 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -69 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -68 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -67 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -66 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -65 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -64 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -63 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -62 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -61 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -60 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -59 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -58 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -57 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -56 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -55 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -54 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -53 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -52 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -51 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -50 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -49 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -48 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -47 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -46 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -45 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -44 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -43 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -42 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -41 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -40 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -39 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -38 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -37 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -36 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -35 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -34 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -33 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -32 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -31 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -30 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -29 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -28 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -27 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -26 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -25 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -24 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -23 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -22 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -21 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -20 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -19 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -18 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -17 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -16 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -15 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -14 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -13 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -12 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -11 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -10 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -9 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -8 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -7 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -6 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -5 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -4 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -3 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -2 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col -1 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 0 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 1 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 2 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 3 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 4 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 5 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 6 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 7 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 8 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 9 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 10 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 11 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 12 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 13 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 14 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 15 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 16 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 17 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 18 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 19 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 20 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 21 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 22 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 23 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 24 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 25 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 26 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 27 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 28 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 29 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 30 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 31 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 32 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 33 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 34 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 35 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 36 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 37 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 38 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 39 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 40 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 41 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 42 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 43 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 44 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 45 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 46 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 47 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 48 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 49 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 50 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 51 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 52 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 53 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 54 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 55 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 56 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 57 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 58 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 59 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 60 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 61 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 62 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 63 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 64 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 65 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 66 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 67 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 68 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 69 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 70 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 71 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 72 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 73 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 74 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 75 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 76 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 77 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 78 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 79 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 80 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 81 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 82 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 83 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 84 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 85 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 86 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 87 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 88 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 89 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 90 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 91 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 92 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 93 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 94 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 95 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 96 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 97 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 98 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 99 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 100 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 101 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 102 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 103 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 104 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 105 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 106 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 107 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 108 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 109 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 110 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 111 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 112 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 113 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 114 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 115 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 116 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 117 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 118 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 119 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 120 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 121 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 122 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 123 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 124 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 125 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 126 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 127 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 128 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 129 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 130 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 131 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 132 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 133 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 134 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 135 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 136 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 137 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 138 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 139 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 140 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 141 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 142 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 143 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 144 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 145 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 146 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 147 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 148 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 149 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 150 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 151 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 152 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 153 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 154 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 155 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 156 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 157 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 158 r,g,b = 0, 0, 42" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101010" report "row 120, col 159 r,g,b = 0, 0, 42" severity error;
		wait until done = '1'; row <= "001110110";
		wait until done = '1'; row <= "001110101";
		wait until done = '1'; row <= "001110100";
		wait until done = '1'; row <= "001110011";
		wait until done = '1'; row <= "001110010";
		wait until done = '1'; row <= "001110001";
		wait until done = '1'; row <= "001110000";
		wait until done = '1'; row <= "001101111";
		wait until done = '1'; row <= "001101110";
		wait until done = '1'; row <= "001101101";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -159 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -158 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -157 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -156 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -155 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -154 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -153 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -152 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -151 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -150 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -149 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -148 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -147 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -146 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -145 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -144 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -143 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -142 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -141 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -140 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -139 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -138 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -137 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -136 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -135 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -134 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -133 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -132 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -131 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -130 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -129 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -128 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -127 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -126 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -125 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -124 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -123 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -122 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -121 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -120 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -119 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -118 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -117 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -116 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -115 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -114 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -113 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -112 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -111 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -110 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -109 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -108 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -107 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -106 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -105 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -104 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -103 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -102 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -101 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -100 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -99 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -98 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -97 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -96 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -95 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -94 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -93 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -92 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -91 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -90 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -89 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -88 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -87 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -86 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -85 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -84 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -83 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -82 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -81 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -80 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -79 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -78 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -77 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -76 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -75 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -74 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -73 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -72 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -71 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -70 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -69 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -68 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -67 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -66 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -65 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -64 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -63 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -62 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -61 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -60 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -59 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -58 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -57 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -56 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -55 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -54 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -53 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -52 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -51 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -50 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -49 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -48 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -47 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -46 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -45 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -44 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -43 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -42 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -41 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -40 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -39 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -38 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -37 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -36 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -35 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -34 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -33 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -32 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -31 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -30 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -29 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -28 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -27 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -26 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -25 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -24 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -23 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -22 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -21 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -20 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -19 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -18 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -17 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -16 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -15 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -14 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -13 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -12 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -11 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -10 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -9 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -8 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -7 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -6 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -5 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -4 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -3 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -2 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col -1 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 0 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 1 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 2 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 3 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 4 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 5 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 6 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 7 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 8 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 9 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 10 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 11 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 12 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 13 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 14 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 15 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 16 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 17 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 18 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 19 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 20 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 21 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 22 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 23 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 24 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 25 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 26 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 27 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 28 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 29 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 30 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 31 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 32 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 33 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 34 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 35 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 36 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 37 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 38 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 39 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 40 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 41 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 42 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 43 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 44 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 45 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 46 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 47 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 48 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 49 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 50 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 51 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 52 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 53 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 54 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 55 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 56 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 57 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 58 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 59 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 60 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 61 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 62 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 63 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 64 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 65 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 66 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 67 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 68 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 69 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 70 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 71 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 72 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 73 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 74 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 75 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 76 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 77 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 78 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 79 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 80 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 81 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 82 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 83 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 84 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 85 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 86 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 87 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 88 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 89 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 90 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 91 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 92 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 93 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 94 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 95 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 96 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 97 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 98 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 99 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 100 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 101 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 102 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 103 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 104 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 105 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 106 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 107 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 108 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 109 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 110 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 111 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 112 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 113 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 114 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 115 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 116 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 117 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 118 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 119 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 120 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 121 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 122 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 123 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 124 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 125 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 126 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 127 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 128 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 129 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 130 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 131 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 132 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 133 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 134 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 135 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 136 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 137 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 138 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 139 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 140 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 141 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 142 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 143 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 144 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 145 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 146 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 147 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 148 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 149 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 150 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 151 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 152 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 153 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 154 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 155 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 156 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 157 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 158 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 110, col 159 r,g,b = 0, 0, 41" severity error;
		wait until done = '1'; row <= "001101100";
		wait until done = '1'; row <= "001101011";
		wait until done = '1'; row <= "001101010";
		wait until done = '1'; row <= "001101001";
		wait until done = '1'; row <= "001101000";
		wait until done = '1'; row <= "001100111";
		wait until done = '1'; row <= "001100110";
		wait until done = '1'; row <= "001100101";
		wait until done = '1'; row <= "001100100";
		wait until done = '1'; row <= "001100011";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -159 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -158 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -157 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -156 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -155 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -154 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -153 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -152 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -151 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -150 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -149 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -148 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -147 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -146 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -145 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -144 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -143 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -142 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -141 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -140 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -139 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -138 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -137 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -136 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -135 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -134 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -133 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -132 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -131 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -130 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -129 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -128 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -127 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -126 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -125 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -124 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -123 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -122 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -121 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -120 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -119 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -118 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -117 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -116 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -115 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -114 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -113 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -112 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -111 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -110 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -109 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -108 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -107 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -106 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -105 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -104 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -103 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -102 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -101 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -100 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -99 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -98 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -97 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -96 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -95 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -94 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -93 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -92 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -91 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -90 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -89 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -88 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -87 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -86 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -85 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -84 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -83 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -82 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -81 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -80 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -79 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -78 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -77 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -76 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -75 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -74 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -73 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -72 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -71 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -70 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -69 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -68 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -67 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -66 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -65 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -64 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -63 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -62 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -61 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -60 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -59 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -58 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -57 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -56 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -55 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -54 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -53 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -52 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -51 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -50 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -49 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -48 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -47 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -46 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -45 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -44 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -43 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -42 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -41 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -40 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -39 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -38 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -37 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -36 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -35 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -34 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -33 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -32 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -31 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -30 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -29 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -28 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -27 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -26 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -25 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -24 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -23 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -22 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -21 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -20 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -19 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -18 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -17 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -16 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -15 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -14 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -13 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -12 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -11 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -10 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -9 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -8 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -7 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -6 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -5 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -4 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -3 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -2 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col -1 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 0 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 1 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 2 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 3 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 4 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 5 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 6 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 7 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 8 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 9 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 10 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 11 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 12 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 13 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 14 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 15 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 16 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 17 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 18 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 19 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 20 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 21 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 22 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 23 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 24 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 25 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 26 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 27 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 28 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 29 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 30 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 31 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 32 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 33 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 34 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 35 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 36 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 37 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 38 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 39 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 40 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 41 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 42 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 43 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 44 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 45 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 46 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 47 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 48 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 49 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 50 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 51 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 52 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 53 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 54 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 55 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 56 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 57 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 58 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 59 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 60 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 61 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 62 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 63 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 64 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 65 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 66 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 67 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 68 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 69 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 70 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 71 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 72 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 73 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 74 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 75 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 76 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 77 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 78 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 79 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 80 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 81 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 82 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 83 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 84 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 85 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 86 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 87 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 88 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 89 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 90 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 91 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 92 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 93 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 94 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 95 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 96 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 97 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 98 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 99 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 100 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 101 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 102 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 103 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 104 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 105 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 106 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 107 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 108 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 109 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 110 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 111 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 112 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 113 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 114 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 115 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 116 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 117 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 118 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 119 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 120 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 121 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 122 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 123 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 124 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 125 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 126 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 127 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 128 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 129 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 130 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 131 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 132 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 133 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 134 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 135 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 136 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 137 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 138 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 139 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 140 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 141 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 142 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 143 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 144 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 145 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 146 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 147 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 148 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 149 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 150 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 151 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 152 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 153 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 154 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 155 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 156 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 157 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 158 r,g,b = 0, 0, 41" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101001" report "row 100, col 159 r,g,b = 0, 0, 41" severity error;
		wait until done = '1'; row <= "001100010";
		wait until done = '1'; row <= "001100001";
		wait until done = '1'; row <= "001100000";
		wait until done = '1'; row <= "001011111";
		wait until done = '1'; row <= "001011110";
		wait until done = '1'; row <= "001011101";
		wait until done = '1'; row <= "001011100";
		wait until done = '1'; row <= "001011011";
		wait until done = '1'; row <= "001011010";
		wait until done = '1'; row <= "001011001";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -159 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -158 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -157 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -156 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -155 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -154 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -153 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -152 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -151 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -150 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -149 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -148 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -147 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -146 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -145 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -144 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -143 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -142 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -141 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -140 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -139 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -138 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -137 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -136 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -135 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -134 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -133 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -132 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -131 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -130 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -129 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -128 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -127 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -126 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -125 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -124 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -123 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -122 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -121 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -120 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -119 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -118 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -117 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -116 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -115 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -114 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -113 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -112 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -111 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -110 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -109 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -108 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -107 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -106 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -105 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -104 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -103 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -102 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -101 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -100 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -99 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -98 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -97 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -96 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -95 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -94 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -93 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -92 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -91 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -90 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -89 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -88 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -87 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -86 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -85 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -84 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -83 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -82 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -81 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -80 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -79 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -78 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -77 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -76 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -75 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -74 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -73 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -72 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -71 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -70 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -69 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -68 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -67 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -66 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -65 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -64 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -63 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -62 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -61 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -60 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -59 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -58 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -57 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -56 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -55 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -54 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -53 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -52 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -51 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -50 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -49 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -48 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -47 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -46 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -45 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -44 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -43 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -42 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -41 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -40 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -39 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -38 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -37 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -36 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -35 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -34 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -33 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -32 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -31 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -30 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -29 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -28 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -27 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -26 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -25 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -24 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -23 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -22 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -21 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -20 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -19 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -18 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -17 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -16 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -15 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -14 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -13 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -12 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -11 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -10 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -9 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -8 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -7 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -6 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -5 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -4 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -3 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -2 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col -1 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 0 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 1 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 2 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 3 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 4 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 5 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 6 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 7 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 8 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 9 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 10 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 11 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 12 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 13 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 14 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 15 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 16 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 17 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 18 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 19 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 20 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 21 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 22 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 23 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 24 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 25 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 26 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 27 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 28 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 29 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 30 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 31 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 32 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 33 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 34 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 35 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 36 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 37 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 38 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 39 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 40 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 41 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 42 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 43 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 44 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 45 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 46 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 47 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 48 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 49 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 50 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 51 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 52 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 53 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 54 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 55 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 56 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 57 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 58 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 59 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 60 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 61 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 62 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 63 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 64 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 65 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 66 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 67 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 68 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 69 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 70 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 71 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 72 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 73 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 74 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 75 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 76 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 77 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 78 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 79 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 80 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 81 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 82 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 83 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 84 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 85 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 86 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 87 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 88 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 89 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 90 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 91 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 92 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 93 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 94 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 95 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 96 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 97 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 98 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 99 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 100 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 101 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 102 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 103 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 104 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 105 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 106 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 107 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 108 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 109 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 110 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 111 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 112 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 113 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 114 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 115 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 116 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 117 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 118 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 119 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 120 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 121 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 122 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 123 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 124 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 125 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 126 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 127 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 128 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 129 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 130 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 131 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 132 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 133 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 134 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 135 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 136 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 137 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 138 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 139 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 140 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 141 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 142 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 143 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 144 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 145 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 146 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 147 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 148 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 149 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 150 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 151 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 152 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 153 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 154 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 155 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 156 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 157 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 158 r,g,b = 0, 0, 40" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "000000000000101000" report "row 90, col 159 r,g,b = 0, 0, 40" severity error;
		wait until done = '1'; row <= "001011000";
		wait until done = '1'; row <= "001010111";
		wait until done = '1'; row <= "001010110";
		wait until done = '1'; row <= "001010101";
		wait until done = '1'; row <= "001010100";
		wait until done = '1'; row <= "001010011";
		wait until done = '1'; row <= "001010010";
		wait until done = '1'; row <= "001010001";
		wait until done = '1'; row <= "001010000";
		wait until done = '1'; row <= "001001111";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -159 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -158 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -157 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -156 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -155 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -154 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -153 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -152 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -151 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -150 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -149 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -148 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -147 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -146 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -145 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -144 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -143 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -142 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -141 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -140 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -139 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -138 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -137 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -136 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -135 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -134 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -133 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -132 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -131 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -130 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -129 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -128 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -127 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -126 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -125 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -124 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -123 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -122 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -121 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -120 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -119 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -118 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -117 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -116 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -115 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -114 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -113 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -112 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -111 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -110 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -109 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -108 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -107 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -106 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -105 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -104 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -103 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -102 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -101 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -100 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -99 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -98 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -97 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -96 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -95 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -94 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -93 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -92 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -91 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -90 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -89 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -88 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -87 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -86 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -85 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -84 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -83 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -82 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -81 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -80 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -79 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -78 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -77 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -76 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -75 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -74 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -73 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -72 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -71 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -70 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -69 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -68 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -67 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -66 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -65 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -64 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -63 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -62 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -61 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -60 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -59 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -58 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -57 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -56 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -55 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -54 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -53 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -52 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -51 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -50 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -49 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -48 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -47 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -46 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -45 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -44 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -43 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -42 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -41 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -40 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -39 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -38 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -37 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -36 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -35 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -34 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -33 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -32 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -31 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -30 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -29 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -28 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -27 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -26 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -25 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -24 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -23 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -22 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -21 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -20 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -19 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -18 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -17 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -16 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -15 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -14 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -13 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -12 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -11 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -10 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -9 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -8 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -7 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -6 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -5 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -4 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -3 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -2 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col -1 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 0 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 1 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 2 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 3 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 4 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 5 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 6 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 7 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 8 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 9 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 10 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 11 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 12 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 13 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 14 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 15 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 16 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 17 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 18 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 19 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 20 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 21 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 22 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 23 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 24 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 25 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 26 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 27 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 28 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 29 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 30 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 31 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 32 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 33 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 34 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 35 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 36 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 37 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 38 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 39 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 40 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 41 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 42 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 43 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 44 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 45 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 46 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 47 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 48 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 49 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 50 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 51 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 52 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 53 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 54 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 55 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 56 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 57 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 58 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 59 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 60 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 61 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 62 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 63 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 64 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 65 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 66 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 67 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 68 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 69 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 70 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 71 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 72 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 73 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 74 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 75 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 76 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 77 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 78 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 79 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 80 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 81 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 82 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 83 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 84 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 85 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 86 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 87 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 88 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 89 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 90 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 91 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 92 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 93 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 94 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 95 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 96 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 97 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 98 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 99 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 100 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 101 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 102 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 103 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 104 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 105 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 106 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 107 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 108 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 109 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 110 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 111 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 112 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 113 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 114 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 115 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 116 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 117 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 118 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 119 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 120 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 121 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 122 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 123 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 124 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 125 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 126 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 127 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 128 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 129 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 130 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 131 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 132 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 133 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 134 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 135 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 136 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 137 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 138 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 139 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 140 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 141 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 142 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 143 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 144 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 145 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 146 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 147 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 148 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 149 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 150 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 151 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 152 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 153 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 154 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 155 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 156 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 157 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 158 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 80, col 159 r,g,b = 0, 0, 39" severity error;
		wait until done = '1'; row <= "001001110";
		wait until done = '1'; row <= "001001101";
		wait until done = '1'; row <= "001001100";
		wait until done = '1'; row <= "001001011";
		wait until done = '1'; row <= "001001010";
		wait until done = '1'; row <= "001001001";
		wait until done = '1'; row <= "001001000";
		wait until done = '1'; row <= "001000111";
		wait until done = '1'; row <= "001000110";
		wait until done = '1'; row <= "001000101";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -159 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -158 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -157 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -156 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -155 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -154 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -153 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -152 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -151 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -150 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -149 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -148 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -147 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -146 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -145 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -144 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -143 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -142 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -141 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -140 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -139 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -138 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -137 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -136 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -135 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -134 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -133 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -132 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -131 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -130 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -129 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -128 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -127 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -126 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -125 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -124 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -123 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -122 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -121 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -120 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -119 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -118 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -117 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -116 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -115 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -114 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -113 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -112 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -111 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -110 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -109 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -108 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -107 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -106 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -105 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -104 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -103 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -102 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -101 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -100 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -99 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -98 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -97 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -96 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -95 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -94 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -93 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -92 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -91 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -90 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -89 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -88 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -87 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -86 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -85 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -84 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -83 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -82 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -81 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -80 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -79 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -78 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -77 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -76 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -75 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -74 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -73 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -72 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -71 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -70 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -69 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -68 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -67 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -66 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -65 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -64 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -63 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -62 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -61 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -60 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -59 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -58 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -57 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -56 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -55 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -54 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -53 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -52 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -51 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -50 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -49 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -48 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -47 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -46 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -45 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -44 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -43 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -42 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -41 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -40 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -39 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -38 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -37 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -36 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -35 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -34 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -33 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -32 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -31 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -30 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -29 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -28 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -27 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -26 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -25 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -24 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -23 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -22 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -21 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -20 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -19 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -18 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -17 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -16 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -15 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -14 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -13 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -12 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -11 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -10 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -9 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -8 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -7 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -6 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -5 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -4 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -3 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -2 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col -1 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 0 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 1 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 2 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 3 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 4 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 5 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 6 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 7 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 8 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 9 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 10 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 11 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 12 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 13 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 14 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 15 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 16 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 17 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 18 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 19 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 20 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 21 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 22 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 23 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 24 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 25 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 26 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 27 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 28 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 29 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 30 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 31 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 32 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 33 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 34 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 35 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 36 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 37 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 38 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 39 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 40 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 41 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 42 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 43 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 44 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 45 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 46 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 47 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 48 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 49 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 50 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 51 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 52 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 53 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 54 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 55 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 56 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 57 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 58 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 59 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 60 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 61 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 62 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 63 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 64 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 65 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 66 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 67 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 68 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 69 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 70 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 71 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 72 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 73 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 74 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 75 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 76 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 77 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 78 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 79 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 80 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 81 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 82 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 83 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 84 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 85 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 86 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 87 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 88 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 89 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 90 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 91 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 92 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 93 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 94 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 95 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 96 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 97 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 98 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 99 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 100 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 101 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 102 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 103 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 104 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 105 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 106 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 107 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 108 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 109 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 110 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 111 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 112 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 113 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 114 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 115 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 116 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 117 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 118 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 119 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 120 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 121 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 122 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 123 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 124 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 125 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 126 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 127 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 128 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 129 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 130 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 131 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 132 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 133 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 134 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 135 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 136 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 137 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 138 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 139 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 140 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 141 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 142 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 143 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 144 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 145 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 146 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 147 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 148 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 149 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 150 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 151 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 152 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 153 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 154 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 155 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 156 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 157 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 158 r,g,b = 0, 0, 39" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100111" report "row 70, col 159 r,g,b = 0, 0, 39" severity error;
		wait until done = '1'; row <= "001000100";
		wait until done = '1'; row <= "001000011";
		wait until done = '1'; row <= "001000010";
		wait until done = '1'; row <= "001000001";
		wait until done = '1'; row <= "001000000";
		wait until done = '1'; row <= "000111111";
		wait until done = '1'; row <= "000111110";
		wait until done = '1'; row <= "000111101";
		wait until done = '1'; row <= "000111100";
		wait until done = '1'; row <= "000111011";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -159 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -158 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -157 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -156 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -155 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -154 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -153 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -152 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -151 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -150 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -149 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -148 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -147 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -146 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -145 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -144 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -143 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -142 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -141 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -140 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -139 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -138 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -137 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -136 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -135 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -134 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -133 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -132 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -131 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -130 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -129 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -128 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -127 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -126 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -125 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -124 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -123 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -122 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -121 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -120 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -119 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -118 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -117 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -116 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -115 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -114 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -113 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -112 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -111 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -110 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -109 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -108 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -107 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -106 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -105 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -104 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -103 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -102 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -101 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -100 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -99 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -98 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -97 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -96 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -95 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -94 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -93 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -92 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -91 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -90 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -89 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -88 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -87 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -86 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -85 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -84 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -83 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -82 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -81 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -80 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -79 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -78 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -77 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -76 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -75 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -74 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -73 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -72 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -71 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -70 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -69 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -68 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -67 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -66 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -65 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -64 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -63 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -62 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -61 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -60 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -59 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -58 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -57 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -56 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -55 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -54 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -53 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -52 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -51 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -50 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -49 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -48 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -47 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -46 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -45 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -44 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -43 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -42 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -41 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -40 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -39 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -38 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -37 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -36 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -35 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -34 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -33 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -32 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -31 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -30 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -29 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -28 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -27 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -26 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -25 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -24 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -23 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -22 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -21 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -20 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -19 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -18 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -17 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -16 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -15 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -14 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -13 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -12 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -11 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -10 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -9 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -8 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -7 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -6 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -5 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -4 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -3 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -2 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col -1 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 0 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 1 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 2 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 3 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 4 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 5 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 6 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 7 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 8 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 9 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 10 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 11 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 12 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 13 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 14 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 15 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 16 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 17 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 18 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 19 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 20 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 21 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 22 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 23 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 24 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 25 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 26 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 27 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 28 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 29 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 30 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 31 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 32 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 33 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 34 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 35 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 36 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 37 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 38 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 39 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 40 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 41 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 42 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 43 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 44 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 45 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 46 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 47 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 48 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 49 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 50 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 51 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 52 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 53 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 54 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 55 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 56 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 57 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 58 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 59 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 60 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 61 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 62 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 63 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 64 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 65 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 66 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 67 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 68 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 69 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 70 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 71 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 72 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 73 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 74 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 75 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 76 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 77 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 78 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 79 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 80 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 81 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 82 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 83 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 84 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 85 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 86 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 87 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 88 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 89 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 90 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 91 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 92 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 93 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 94 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 95 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 96 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 97 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 98 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 99 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 100 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 101 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 102 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 103 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 104 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 105 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 106 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 107 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 108 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 109 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 110 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 111 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 112 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 113 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 114 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 115 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 116 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 117 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 118 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 119 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 120 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 121 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 122 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 123 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 124 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 125 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 126 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 127 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 128 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 129 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 130 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 131 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 132 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 133 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 134 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 135 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 136 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 137 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 138 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 139 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 140 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 141 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 142 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 143 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 144 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 145 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 146 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 147 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 148 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 149 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 150 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 151 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 152 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 153 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 154 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 155 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 156 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 157 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 158 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 60, col 159 r,g,b = 0, 0, 38" severity error;
		wait until done = '1'; row <= "000111010";
		wait until done = '1'; row <= "000111001";
		wait until done = '1'; row <= "000111000";
		wait until done = '1'; row <= "000110111";
		wait until done = '1'; row <= "000110110";
		wait until done = '1'; row <= "000110101";
		wait until done = '1'; row <= "000110100";
		wait until done = '1'; row <= "000110011";
		wait until done = '1'; row <= "000110010";
		wait until done = '1'; row <= "000110001";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -159 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -158 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -157 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -156 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -155 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -154 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -153 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -152 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -151 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -150 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -149 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -148 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -147 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -146 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -145 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -144 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -143 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -142 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -141 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -140 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -139 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -138 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -137 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -136 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -135 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -134 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -133 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -132 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -131 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -130 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -129 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -128 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -127 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -126 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -125 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -124 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -123 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -122 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -121 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -120 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -119 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -118 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -117 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -116 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -115 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -114 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -113 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -112 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -111 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -110 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -109 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -108 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -107 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -106 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -105 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -104 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -103 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -102 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -101 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -100 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -99 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -98 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -97 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -96 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -95 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -94 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -93 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -92 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -91 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -90 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -89 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -88 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -87 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -86 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -85 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -84 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -83 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -82 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -81 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -80 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -79 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -78 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -77 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -76 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -75 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -74 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -73 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -72 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -71 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -70 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -69 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -68 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -67 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -66 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -65 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -64 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -63 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -62 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -61 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -60 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -59 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -58 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -57 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -56 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -55 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -54 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -53 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -52 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -51 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col -50 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -49 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -48 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -47 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -46 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -45 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -44 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -43 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -42 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -41 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -40 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -39 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -38 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -37 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -36 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -35 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -34 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -33 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -32 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -31 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -30 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -29 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -28 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -27 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -26 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -25 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -24 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -23 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -22 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -21 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -20 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -19 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -18 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -17 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -16 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -15 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -14 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -13 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -12 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -11 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -10 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -9 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -8 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -7 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -6 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -5 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -4 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -3 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -2 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col -1 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 0 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 1 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 2 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 3 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 4 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 5 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 6 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 7 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 8 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 9 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 10 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 11 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 12 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 13 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 14 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 15 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 16 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 17 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 18 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 19 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 20 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 21 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 22 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 23 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 24 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 25 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 26 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 27 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 28 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 29 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 30 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 31 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 32 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 33 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 34 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 35 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 36 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 37 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 38 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 39 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 40 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 41 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 42 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 43 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 44 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 45 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 46 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 47 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 48 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 50, col 49 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 50 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 51 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 52 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 53 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 54 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 55 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 56 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 57 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 58 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 59 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 60 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 61 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 62 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 63 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 64 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 65 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 66 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 67 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 68 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 69 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 70 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 71 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 72 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 73 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 74 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 75 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 76 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 77 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 78 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 79 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 80 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 81 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 82 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 83 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 84 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 85 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 86 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 87 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 88 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 89 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 90 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 91 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 92 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 93 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 94 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 95 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 96 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 97 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 98 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 99 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 100 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 101 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 102 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 103 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 104 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 105 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 106 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 107 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 108 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 109 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 110 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 111 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 112 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 113 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 114 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 115 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 116 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 117 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 118 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 119 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 120 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 121 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 122 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 123 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 124 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 125 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 126 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 127 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 128 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 129 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 130 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 131 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 132 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 133 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 134 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 135 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 136 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 137 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 138 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 139 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 140 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 141 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 142 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 143 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 144 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 145 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 146 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 147 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 148 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 149 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 150 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 151 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 152 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 153 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 154 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 155 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 156 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 157 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 158 r,g,b = 0, 0, 38" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100110" report "row 50, col 159 r,g,b = 0, 0, 38" severity error;
		wait until done = '1'; row <= "000110000";
		wait until done = '1'; row <= "000101111";
		wait until done = '1'; row <= "000101110";
		wait until done = '1'; row <= "000101101";
		wait until done = '1'; row <= "000101100";
		wait until done = '1'; row <= "000101011";
		wait until done = '1'; row <= "000101010";
		wait until done = '1'; row <= "000101001";
		wait until done = '1'; row <= "000101000";
		wait until done = '1'; row <= "000100111";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -159 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -158 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -157 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -156 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -155 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -154 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -153 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -152 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -151 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -150 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -149 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -148 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -147 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -146 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -145 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -144 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -143 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -142 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -141 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -140 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -139 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -138 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -137 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -136 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -135 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -134 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -133 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -132 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -131 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -130 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -129 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -128 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -127 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -126 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -125 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -124 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -123 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -122 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -121 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -120 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -119 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -118 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -117 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -116 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -115 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -114 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -113 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -112 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -111 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -110 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -109 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -108 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -107 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -106 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -105 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -104 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -103 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -102 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -101 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -100 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -99 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -98 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -97 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -96 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -95 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -94 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -93 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -92 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -91 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -90 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -89 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -88 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -87 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -86 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -85 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -84 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -83 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -82 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -81 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -80 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -79 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -78 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -77 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -76 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -75 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -74 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -73 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -72 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -71 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -70 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -69 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -68 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -67 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -66 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -65 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -64 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -63 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -62 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -61 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -60 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -59 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -58 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -57 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -56 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -55 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -54 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -53 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -52 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -51 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -50 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -49 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -48 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -47 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -46 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -45 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -44 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -43 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -42 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -41 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col -40 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -39 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -38 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -37 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -36 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -35 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -34 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -33 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -32 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -31 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -30 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -29 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -28 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -27 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -26 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -25 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -24 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -23 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -22 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -21 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -20 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -19 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -18 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -17 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -16 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -15 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -14 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -13 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -12 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -11 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -10 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -9 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -8 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -7 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -6 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -5 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -4 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -3 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -2 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col -1 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 0 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 1 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 2 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 3 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 4 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 5 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 6 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 7 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 8 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 9 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 10 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 11 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 12 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 13 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 14 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 15 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 16 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 17 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 18 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 19 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 20 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 21 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 22 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 23 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 24 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 25 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 26 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 27 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 28 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 29 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 30 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 31 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 32 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 33 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 34 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 35 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 36 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 37 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 38 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 40, col 39 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 40 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 41 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 42 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 43 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 44 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 45 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 46 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 47 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 48 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 49 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 50 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 51 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 52 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 53 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 54 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 55 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 56 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 57 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 58 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 59 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 60 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 61 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 62 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 63 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 64 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 65 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 66 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 67 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 68 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 69 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 70 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 71 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 72 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 73 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 74 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 75 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 76 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 77 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 78 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 79 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 80 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 81 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 82 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 83 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 84 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 85 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 86 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 87 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 88 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 89 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 90 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 91 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 92 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 93 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 94 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 95 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 96 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 97 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 98 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 99 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 100 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 101 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 102 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 103 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 104 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 105 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 106 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 107 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 108 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 109 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 110 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 111 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 112 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 113 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 114 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 115 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 116 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 117 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 118 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 119 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 120 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 121 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 122 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 123 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 124 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 125 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 126 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 127 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 128 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 129 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 130 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 131 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 132 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 133 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 134 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 135 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 136 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 137 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 138 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 139 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 140 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 141 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 142 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 143 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 144 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 145 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 146 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 147 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 148 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 149 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 150 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 151 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 152 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 153 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 154 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 155 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 156 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 157 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 158 r,g,b = 0, 0, 37" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100101" report "row 40, col 159 r,g,b = 0, 0, 37" severity error;
		wait until done = '1'; row <= "000100110";
		wait until done = '1'; row <= "000100101";
		wait until done = '1'; row <= "000100100";
		wait until done = '1'; row <= "000100011";
		wait until done = '1'; row <= "000100010";
		wait until done = '1'; row <= "000100001";
		wait until done = '1'; row <= "000100000";
		wait until done = '1'; row <= "000011111";
		wait until done = '1'; row <= "000011110";
		wait until done = '1'; row <= "000011101";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -159 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -158 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -157 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -156 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -155 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -154 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -153 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -152 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -151 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -150 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -149 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -148 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -147 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -146 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -145 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -144 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -143 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -142 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -141 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -140 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -139 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -138 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -137 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -136 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -135 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -134 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -133 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -132 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -131 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -130 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -129 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -128 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -127 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -126 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -125 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -124 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -123 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -122 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -121 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -120 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -119 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -118 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -117 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -116 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -115 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -114 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -113 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -112 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -111 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -110 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -109 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -108 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -107 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -106 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -105 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -104 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -103 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -102 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -101 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -100 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -99 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -98 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -97 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -96 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -95 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -94 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -93 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -92 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -91 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -90 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -89 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -88 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -87 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -86 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -85 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -84 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -83 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -82 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -81 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -80 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -79 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -78 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -77 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -76 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -75 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -74 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -73 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -72 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -71 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -70 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -69 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -68 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -67 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -66 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -65 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -64 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -63 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -62 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -61 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -60 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -59 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -58 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -57 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -56 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -55 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -54 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -53 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -52 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -51 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -50 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -49 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -48 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -47 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -46 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -45 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -44 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -43 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -42 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -41 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -40 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -39 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -38 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -37 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -36 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -35 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -34 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -33 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -32 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -31 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col -30 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -29 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -28 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -27 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -26 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -25 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -24 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -23 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -22 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -21 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -20 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -19 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -18 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -17 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -16 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -15 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -14 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -13 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -12 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -11 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -10 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -9 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -8 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -7 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -6 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -5 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -4 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -3 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -2 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col -1 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 0 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 1 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 2 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 3 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 4 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 5 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 6 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 7 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 8 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 9 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 10 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 11 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 12 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 13 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 14 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 15 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 16 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 17 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 18 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 19 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 20 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 21 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 22 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 23 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 24 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 25 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 26 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 27 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 28 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row 30, col 29 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 30 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 31 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 32 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 33 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 34 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 35 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 36 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 37 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 38 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 39 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 40 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 41 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 42 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 43 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 44 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 45 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 46 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 47 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 48 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 49 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 50 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 51 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 52 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 53 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 54 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 55 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 56 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 57 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 58 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 59 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 60 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 61 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 62 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 63 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 64 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 65 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 66 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 67 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 68 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 69 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 70 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 71 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 72 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 73 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 74 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 75 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 76 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 77 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 78 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 79 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 80 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 81 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 82 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 83 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 84 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 85 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 86 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 87 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 88 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 89 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 90 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 91 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 92 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 93 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 94 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 95 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 96 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 97 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 98 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 99 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 100 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 101 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 102 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 103 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 104 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 105 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 106 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 107 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 108 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 109 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 110 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 111 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 112 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 113 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 114 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 115 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 116 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 117 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 118 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 119 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 120 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 121 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 122 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 123 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 124 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 125 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 126 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 127 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 128 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 129 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 130 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 131 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 132 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 133 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 134 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 135 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 136 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 137 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 138 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 139 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 140 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 141 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 142 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 143 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 144 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 145 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 146 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 147 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 148 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 149 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 150 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 151 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 152 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 153 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 154 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 155 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 156 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 157 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 158 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 30, col 159 r,g,b = 0, 0, 36" severity error;
		wait until done = '1'; row <= "000011100";
		wait until done = '1'; row <= "000011011";
		wait until done = '1'; row <= "000011010";
		wait until done = '1'; row <= "000011001";
		wait until done = '1'; row <= "000011000";
		wait until done = '1'; row <= "000010111";
		wait until done = '1'; row <= "000010110";
		wait until done = '1'; row <= "000010101";
		wait until done = '1'; row <= "000010100";
		wait until done = '1'; row <= "000010011";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -159 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -158 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -157 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -156 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -155 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -154 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -153 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -152 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -151 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -150 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -149 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -148 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -147 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -146 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -145 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -144 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -143 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -142 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -141 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -140 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -139 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -138 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -137 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -136 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -135 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -134 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -133 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -132 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -131 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -130 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -129 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -128 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -127 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -126 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -125 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -124 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -123 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -122 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -121 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -120 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -119 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -118 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -117 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -116 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -115 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -114 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -113 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -112 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -111 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -110 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -109 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -108 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -107 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -106 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -105 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -104 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -103 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -102 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -101 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -100 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -99 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -98 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -97 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -96 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -95 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -94 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -93 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -92 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -91 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -90 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -89 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -88 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -87 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -86 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -85 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -84 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -83 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -82 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -81 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -80 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -79 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -78 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -77 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -76 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -75 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -74 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -73 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -72 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -71 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -70 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -69 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -68 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -67 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -66 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -65 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -64 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -63 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -62 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -61 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -60 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -59 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -58 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -57 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -56 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -55 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -54 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -53 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -52 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -51 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -50 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -49 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -48 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -47 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -46 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -45 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -44 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -43 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -42 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -41 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -40 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -39 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -38 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -37 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -36 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -35 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -34 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -33 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -32 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -31 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -30 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -29 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -28 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -27 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -26 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -25 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -24 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -23 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -22 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -21 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col -20 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col -19 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col -18 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col -17 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col -16 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col -15 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col -14 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col -13 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col -12 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col -11 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col -10 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col -9 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col -8 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col -7 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col -6 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col -5 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col -4 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col -3 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col -2 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col -1 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col 0 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col 1 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col 2 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col 3 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col 4 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col 5 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col 6 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col 7 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col 8 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col 9 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col 10 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col 11 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col 12 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col 13 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col 14 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col 15 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col 16 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col 17 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col 18 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "000000001111011111" report "row 20, col 19 r,g,b = 0, 15, 31" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 20 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 21 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 22 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 23 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 24 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 25 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 26 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 27 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 28 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 29 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 30 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 31 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 32 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 33 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 34 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 35 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 36 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 37 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 38 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 39 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 40 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 41 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 42 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 43 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 44 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 45 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 46 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 47 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 48 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 49 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 50 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 51 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 52 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 53 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 54 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 55 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 56 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 57 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 58 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 59 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 60 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 61 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 62 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 63 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 64 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 65 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 66 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 67 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 68 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 69 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 70 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 71 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 72 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 73 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 74 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 75 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 76 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 77 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 78 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 79 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 80 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 81 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 82 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 83 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 84 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 85 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 86 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 87 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 88 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 89 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 90 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 91 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 92 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 93 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 94 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 95 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 96 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 97 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 98 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 99 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 100 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 101 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 102 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 103 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 104 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 105 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 106 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 107 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 108 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 109 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 110 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 111 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 112 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 113 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 114 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 115 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 116 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 117 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 118 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 119 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 120 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 121 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 122 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 123 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 124 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 125 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 126 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 127 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 128 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 129 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 130 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 131 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 132 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 133 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 134 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 135 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 136 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 137 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 138 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 139 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 140 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 141 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 142 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 143 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 144 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 145 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 146 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 147 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 148 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 149 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 150 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 151 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 152 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 153 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 154 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 155 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 156 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 157 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 158 r,g,b = 0, 0, 36" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100100" report "row 20, col 159 r,g,b = 0, 0, 36" severity error;
		wait until done = '1'; row <= "000010010";
		wait until done = '1'; row <= "000010001";
		wait until done = '1'; row <= "000010000";
		wait until done = '1'; row <= "000001111";
		wait until done = '1'; row <= "000001110";
		wait until done = '1'; row <= "000001101";
		wait until done = '1'; row <= "000001100";
		wait until done = '1'; row <= "000001011";
		wait until done = '1'; row <= "000001010";
		wait until done = '1'; row <= "000001001";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -159 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -158 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -157 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -156 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -155 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -154 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -153 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -152 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -151 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -150 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -149 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -148 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -147 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -146 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -145 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -144 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -143 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -142 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -141 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -140 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -139 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -138 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -137 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -136 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -135 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -134 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -133 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -132 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -131 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -130 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -129 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -128 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -127 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -126 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -125 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -124 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -123 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -122 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -121 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -120 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -119 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -118 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -117 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -116 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -115 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -114 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -113 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -112 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -111 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -110 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -109 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -108 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -107 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -106 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -105 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -104 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -103 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -102 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -101 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -100 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -99 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -98 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -97 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -96 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -95 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -94 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -93 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -92 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -91 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -90 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -89 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -88 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -87 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -86 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -85 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -84 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -83 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -82 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -81 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -80 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -79 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -78 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -77 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -76 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -75 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -74 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -73 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -72 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -71 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -70 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -69 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -68 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -67 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -66 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -65 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -64 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -63 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -62 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -61 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -60 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -59 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -58 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -57 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -56 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -55 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -54 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -53 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -52 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -51 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -50 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -49 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -48 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -47 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -46 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -45 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -44 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -43 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -42 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -41 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -40 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -39 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -38 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -37 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -36 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -35 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -34 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -33 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -32 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -31 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -30 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -29 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -28 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -27 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -26 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -25 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -24 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -23 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -22 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -21 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -20 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -19 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -18 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -17 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -16 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -15 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -14 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -13 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -12 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -11 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -10 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -9 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -8 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -7 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -6 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -5 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -4 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -3 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -2 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col -1 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 0 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 1 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 2 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 3 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 4 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 5 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 6 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 7 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 8 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 9 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 10 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 11 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 12 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 13 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 14 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 15 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 16 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 17 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 18 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 19 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 20 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 21 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 22 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 23 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 24 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 25 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 26 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 27 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 28 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 29 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 30 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 31 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 32 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 33 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 34 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 35 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 36 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 37 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 38 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 39 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 40 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 41 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 42 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 43 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 44 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 45 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 46 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 47 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 48 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 49 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 50 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 51 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 52 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 53 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 54 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 55 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 56 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 57 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 58 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 59 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 60 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 61 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 62 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 63 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 64 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 65 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 66 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 67 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 68 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 69 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 70 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 71 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 72 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 73 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 74 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 75 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 76 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 77 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 78 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 79 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 80 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 81 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 82 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 83 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 84 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 85 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 86 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 87 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 88 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 89 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 90 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 91 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 92 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 93 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 94 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 95 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 96 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 97 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 98 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 99 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 100 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 101 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 102 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 103 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 104 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 105 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 106 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 107 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 108 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 109 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 110 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 111 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 112 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 113 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 114 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 115 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 116 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 117 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 118 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 119 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 120 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 121 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 122 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 123 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 124 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 125 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 126 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 127 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 128 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 129 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 130 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 131 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 132 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 133 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 134 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 135 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 136 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 137 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 138 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 139 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 140 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 141 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 142 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 143 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 144 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 145 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 146 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 147 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 148 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 149 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 150 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 151 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 152 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 153 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 154 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 155 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 156 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 157 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 158 r,g,b = 0, 0, 35" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100011" report "row 10, col 159 r,g,b = 0, 0, 35" severity error;
		wait until done = '1'; row <= "000001000";
		wait until done = '1'; row <= "000000111";
		wait until done = '1'; row <= "000000110";
		wait until done = '1'; row <= "000000101";
		wait until done = '1'; row <= "000000100";
		wait until done = '1'; row <= "000000011";
		wait until done = '1'; row <= "000000010";
		wait until done = '1'; row <= "000000001";
		wait until done = '1'; row <= "000000000";
		wait until done = '1'; row <= "111111111";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -159 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -158 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -157 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -156 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -155 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -154 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -153 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -152 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -151 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -150 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -149 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -148 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -147 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -146 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -145 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -144 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -143 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -142 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -141 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -140 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -139 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -138 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -137 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -136 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -135 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -134 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -133 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -132 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -131 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -130 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -129 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -128 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -127 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -126 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -125 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -124 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -123 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -122 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -121 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -120 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -119 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -118 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -117 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -116 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -115 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -114 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -113 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -112 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -111 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -110 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -109 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -108 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -107 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -106 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -105 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -104 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -103 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -102 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -101 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -100 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -99 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -98 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -97 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -96 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -95 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -94 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -93 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -92 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -91 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -90 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -89 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -88 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -87 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -86 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -85 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -84 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -83 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -82 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -81 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -80 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -79 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -78 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -77 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -76 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -75 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -74 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -73 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -72 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -71 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -70 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -69 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -68 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -67 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -66 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -65 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -64 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -63 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -62 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -61 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -60 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -59 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -58 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -57 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -56 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -55 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -54 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -53 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -52 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -51 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -50 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -49 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -48 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -47 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -46 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -45 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -44 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -43 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -42 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -41 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -40 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -39 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -38 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -37 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -36 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -35 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -34 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -33 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -32 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -31 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -30 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -29 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -28 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -27 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -26 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -25 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -24 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -23 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -22 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -21 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -20 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -19 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -18 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -17 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -16 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -15 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -14 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -13 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -12 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -11 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -10 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -9 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -8 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -7 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -6 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -5 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -4 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -3 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -2 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col -1 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 0 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 1 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 2 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 3 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 4 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 5 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 6 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 7 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 8 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 9 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 10 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 11 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 12 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 13 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 14 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 15 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 16 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 17 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 18 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 19 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 20 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 21 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 22 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 23 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 24 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 25 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 26 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 27 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 28 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 29 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 30 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 31 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 32 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 33 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 34 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 35 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 36 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 37 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 38 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 39 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 40 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 41 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 42 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 43 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 44 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 45 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 46 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 47 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 48 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 49 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 50 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 51 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 52 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 53 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 54 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 55 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 56 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 57 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 58 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 59 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 60 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 61 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 62 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 63 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 64 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 65 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 66 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 67 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 68 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 69 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 70 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 71 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 72 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 73 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 74 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 75 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 76 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 77 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 78 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 79 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 80 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 81 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 82 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 83 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 84 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 85 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 86 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 87 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 88 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 89 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 90 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 91 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 92 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 93 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 94 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 95 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 96 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 97 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 98 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 99 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 100 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 101 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 102 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 103 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 104 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 105 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 106 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 107 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 108 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 109 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 110 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 111 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 112 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 113 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 114 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 115 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 116 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 117 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 118 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 119 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 120 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 121 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 122 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 123 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 124 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 125 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 126 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 127 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 128 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 129 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 130 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 131 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 132 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 133 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 134 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 135 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 136 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 137 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 138 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 139 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 140 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 141 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 142 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 143 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 144 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 145 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 146 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 147 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 148 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 149 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 150 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 151 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 152 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 153 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 154 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 155 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 156 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 157 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 158 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row 0, col 159 r,g,b = 0, 0, 34" severity error;
		wait until done = '1'; row <= "111111110";
		wait until done = '1'; row <= "111111101";
		wait until done = '1'; row <= "111111100";
		wait until done = '1'; row <= "111111011";
		wait until done = '1'; row <= "111111010";
		wait until done = '1'; row <= "111111001";
		wait until done = '1'; row <= "111111000";
		wait until done = '1'; row <= "111110111";
		wait until done = '1'; row <= "111110110";
		wait until done = '1'; row <= "111110101";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -159 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -158 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -157 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -156 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -155 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -154 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -153 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -152 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -151 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -150 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -149 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -148 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -147 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -146 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -145 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -144 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -143 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -142 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -141 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -140 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -139 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -138 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -137 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -136 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -135 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -134 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -133 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -132 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -131 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -130 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -129 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -128 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -127 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -126 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -125 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -124 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -123 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -122 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -121 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -120 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -119 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -118 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -117 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -116 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -115 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -114 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -113 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -112 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -111 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -110 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -109 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -108 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -107 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -106 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -105 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -104 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -103 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -102 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -101 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -100 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -99 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -98 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -97 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -96 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -95 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -94 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -93 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -92 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -91 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -90 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -89 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -88 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -87 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -86 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -85 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -84 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -83 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -82 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -81 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -80 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -79 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -78 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -77 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -76 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -75 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -74 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -73 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -72 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -71 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -70 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -69 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -68 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -67 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -66 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -65 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -64 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -63 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -62 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -61 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -60 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -59 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -58 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -57 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -56 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -55 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -54 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -53 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -52 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -51 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -50 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -49 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -48 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -47 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -46 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -45 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -44 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -43 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -42 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -41 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -40 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -39 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -38 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -37 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -36 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -35 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -34 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -33 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -32 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -31 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -30 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -29 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -28 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -27 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -26 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -25 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -24 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -23 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -22 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -21 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -20 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -19 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -18 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -17 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -16 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -15 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -14 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -13 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -12 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -11 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -10 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -9 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -8 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -7 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -6 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -5 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -4 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -3 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -2 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col -1 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 0 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 1 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 2 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 3 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 4 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 5 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 6 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 7 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 8 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 9 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 10 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 11 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 12 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 13 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 14 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 15 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 16 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 17 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 18 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 19 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 20 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 21 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 22 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 23 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 24 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 25 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 26 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 27 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 28 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 29 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 30 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 31 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 32 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 33 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 34 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 35 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 36 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 37 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 38 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 39 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 40 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 41 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 42 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 43 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 44 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 45 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 46 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 47 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 48 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 49 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 50 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 51 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 52 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 53 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 54 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 55 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 56 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 57 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 58 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 59 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 60 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 61 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 62 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 63 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 64 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 65 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 66 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 67 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 68 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 69 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 70 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 71 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 72 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 73 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 74 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 75 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 76 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 77 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 78 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -10, col 79 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 80 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 81 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 82 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 83 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 84 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 85 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 86 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 87 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 88 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 89 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 90 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 91 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 92 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 93 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 94 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 95 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 96 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 97 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 98 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 99 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 100 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 101 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 102 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 103 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 104 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 105 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 106 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 107 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 108 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 109 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 110 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 111 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 112 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 113 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 114 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 115 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 116 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 117 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 118 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 119 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 120 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 121 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 122 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 123 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 124 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 125 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 126 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 127 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 128 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 129 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 130 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 131 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 132 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 133 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 134 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 135 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 136 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 137 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 138 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 139 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 140 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 141 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 142 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 143 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 144 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 145 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 146 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 147 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 148 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 149 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 150 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 151 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 152 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 153 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 154 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 155 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 156 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 157 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 158 r,g,b = 0, 0, 34" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "000000000000100010" report "row -10, col 159 r,g,b = 0, 0, 34" severity error;
		wait until done = '1'; row <= "111110100";
		wait until done = '1'; row <= "111110011";
		wait until done = '1'; row <= "111110010";
		wait until done = '1'; row <= "111110001";
		wait until done = '1'; row <= "111110000";
		wait until done = '1'; row <= "111101111";
		wait until done = '1'; row <= "111101110";
		wait until done = '1'; row <= "111101101";
		wait until done = '1'; row <= "111101100";
		wait until done = '1'; row <= "111101011";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -159 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -158 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -157 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -156 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -155 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -154 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -153 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -152 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -151 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -150 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -149 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -148 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -147 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -146 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -145 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -144 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -143 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -142 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -141 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -140 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -139 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -138 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -137 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -136 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -135 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -134 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -133 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -132 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -131 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -130 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -129 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -128 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -127 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -126 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -125 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -124 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -123 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -122 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -121 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -120 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -119 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -118 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -117 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -116 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -115 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -114 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -113 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -112 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -111 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -110 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -109 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -108 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -107 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -106 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -105 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -104 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -103 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -102 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -101 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -100 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -99 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -98 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -97 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -96 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -95 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -94 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -93 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -92 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -91 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -90 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -89 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -88 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -87 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -86 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -85 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -84 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -83 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -82 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -81 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -80 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -79 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -78 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -77 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -76 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -75 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -74 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -73 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -72 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -71 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -70 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -69 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -68 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -67 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -66 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -65 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -64 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -63 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -62 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -61 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -60 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -59 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -58 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -57 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -56 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -55 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -54 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -53 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -52 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -51 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -50 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -49 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -48 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -47 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -46 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -45 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -44 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -43 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -42 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -41 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -40 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -39 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -38 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -37 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -36 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -35 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -34 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -33 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -32 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -31 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -30 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -29 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -28 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -27 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -26 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -25 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -24 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -23 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -22 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -21 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -20 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -19 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -18 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -17 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -16 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -15 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -14 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -13 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -12 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -11 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -10 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -9 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -8 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -7 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -6 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -5 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -4 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -3 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -2 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col -1 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 0 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 1 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 2 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 3 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 4 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 5 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 6 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 7 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 8 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 9 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 10 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 11 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 12 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 13 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 14 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 15 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 16 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 17 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 18 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 19 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 20 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 21 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 22 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 23 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 24 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 25 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 26 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 27 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 28 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 29 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 30 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 31 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 32 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 33 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 34 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 35 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 36 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 37 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 38 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 39 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 40 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 41 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 42 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 43 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 44 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 45 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 46 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 47 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 48 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 49 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 50 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 51 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 52 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 53 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 54 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 55 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 56 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 57 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 58 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 59 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 60 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 61 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 62 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 63 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 64 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 65 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 66 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 67 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 68 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 69 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 70 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 71 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 72 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 73 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 74 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 75 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 76 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 77 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 78 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 79 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 80 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 81 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 82 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 83 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 84 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 85 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 86 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 87 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 88 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 89 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 90 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 91 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 92 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 93 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 94 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 95 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 96 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 97 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 98 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 99 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 100 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 101 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 102 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 103 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 104 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 105 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 106 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 107 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 108 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 109 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 110 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 111 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 112 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 113 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 114 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 115 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 116 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 117 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 118 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 119 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 120 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 121 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 122 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 123 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 124 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 125 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 126 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 127 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 128 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 129 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 130 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 131 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 132 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 133 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 134 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 135 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 136 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 137 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 138 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 139 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 140 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 141 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 142 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 143 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 144 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 145 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 146 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 147 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 148 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 149 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 150 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 151 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 152 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 153 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 154 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 155 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 156 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 157 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 158 r,g,b = 15, 15, 31" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "001111001111011111" report "row -20, col 159 r,g,b = 15, 15, 31" severity error;
		wait until done = '1'; row <= "111101010";
		wait until done = '1'; row <= "111101001";
		wait until done = '1'; row <= "111101000";
		wait until done = '1'; row <= "111100111";
		wait until done = '1'; row <= "111100110";
		wait until done = '1'; row <= "111100101";
		wait until done = '1'; row <= "111100100";
		wait until done = '1'; row <= "111100011";
		wait until done = '1'; row <= "111100010";
		wait until done = '1'; row <= "111100001";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -159 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -158 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -157 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -156 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -155 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -154 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -153 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -152 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -151 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -150 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -149 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -148 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -147 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -146 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -145 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -144 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -143 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -142 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -141 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -140 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -139 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -138 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -137 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -136 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -135 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -134 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -133 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -132 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -131 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -130 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -129 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -128 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -127 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -126 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -125 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -124 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -123 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -122 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -121 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -120 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -119 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -118 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -117 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -116 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -115 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -114 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -113 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -112 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -111 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -110 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -109 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -108 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -107 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -106 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -105 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -104 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -103 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -102 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -101 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -100 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -99 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -98 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -97 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -96 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -95 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -94 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -93 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -92 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -91 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -90 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -89 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -88 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -87 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -86 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -85 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -84 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -83 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -82 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -81 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -80 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -79 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -78 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -77 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -76 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -75 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -74 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -73 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -72 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -71 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -70 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -69 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -68 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -67 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -66 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -65 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -64 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -63 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -62 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -61 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -60 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -59 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -58 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -57 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -56 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -55 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -54 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -53 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -52 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -51 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -50 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -49 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -48 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -47 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -46 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -45 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -44 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -43 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -42 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -41 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -40 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -39 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -38 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col -37 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -36 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -35 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -34 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -33 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -32 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -31 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -30 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -29 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -28 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -27 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -26 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -25 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -24 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -23 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -22 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -21 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -20 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -19 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -18 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -17 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -16 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -15 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -14 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -13 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -12 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -11 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -10 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -9 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -8 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -7 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -6 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -5 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -4 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -3 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -2 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col -1 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 0 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 1 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 2 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 3 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 4 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 5 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 6 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 7 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 8 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 9 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 10 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 11 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 12 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 13 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 14 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 15 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 16 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 17 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 18 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 19 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 20 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 21 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 22 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 23 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 24 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 25 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 26 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 27 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 28 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 29 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 30 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 31 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 32 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 33 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 34 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 35 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -30, col 36 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 37 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 38 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 39 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 40 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 41 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 42 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 43 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 44 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 45 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 46 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 47 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 48 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 49 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 50 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 51 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 52 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 53 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 54 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 55 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 56 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 57 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 58 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 59 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 60 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 61 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 62 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 63 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 64 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 65 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 66 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 67 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 68 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 69 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 70 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 71 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 72 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 73 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 74 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 75 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 76 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 77 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 78 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 79 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 80 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 81 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 82 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 83 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 84 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 85 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 86 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 87 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 88 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 89 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 90 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 91 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 92 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 93 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 94 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 95 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 96 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 97 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 98 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 99 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 100 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 101 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 102 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 103 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 104 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 105 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 106 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 107 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 108 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 109 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 110 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 111 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 112 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 113 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 114 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 115 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 116 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 117 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 118 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 119 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 120 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 121 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 122 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 123 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 124 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 125 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 126 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 127 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 128 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 129 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 130 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 131 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 132 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 133 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 134 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 135 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 136 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 137 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 138 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 139 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 140 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 141 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 142 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 143 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 144 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 145 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 146 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 147 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 148 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 149 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 150 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 151 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 152 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 153 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 154 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 155 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 156 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 157 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 158 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -30, col 159 r,g,b = 31, 31, 15" severity error;
		wait until done = '1'; row <= "111100000";
		wait until done = '1'; row <= "111011111";
		wait until done = '1'; row <= "111011110";
		wait until done = '1'; row <= "111011101";
		wait until done = '1'; row <= "111011100";
		wait until done = '1'; row <= "111011011";
		wait until done = '1'; row <= "111011010";
		wait until done = '1'; row <= "111011001";
		wait until done = '1'; row <= "111011000";
		wait until done = '1'; row <= "111010111";
		screen_col_to_read <= "000000000"; wait for 40 ns;
		screen_col_to_read <= "000000001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -159 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000000010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -158 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000000011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -157 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000000100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -156 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000000101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -155 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000000110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -154 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000000111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -153 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000001000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -152 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000001001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -151 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000001010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -150 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000001011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -149 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000001100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -148 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000001101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -147 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000001110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -146 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000001111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -145 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000010000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -144 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000010001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -143 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000010010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -142 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000010011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -141 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000010100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -140 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000010101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -139 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000010110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -138 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000010111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -137 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000011000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -136 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000011001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -135 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000011010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -134 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000011011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -133 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000011100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -132 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000011101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -131 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000011110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -130 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000011111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -129 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000100000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -128 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000100001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -127 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000100010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -126 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000100011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -125 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000100100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -124 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000100101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -123 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000100110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -122 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000100111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -121 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000101000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -120 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000101001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -119 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000101010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -118 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000101011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -117 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000101100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -116 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000101101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -115 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000101110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -114 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000101111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -113 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000110000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -112 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000110001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -111 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000110010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -110 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000110011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -109 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000110100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -108 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000110101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -107 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000110110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -106 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000110111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -105 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000111000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -104 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000111001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -103 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000111010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -102 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000111011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -101 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000111100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -100 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000111101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -99 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000111110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -98 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "000111111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -97 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001000000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -96 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001000001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -95 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001000010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -94 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001000011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -93 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001000100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -92 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001000101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -91 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001000110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -90 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001000111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -89 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001001000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -88 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001001001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -87 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001001010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -86 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001001011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -85 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001001100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -84 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001001101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -83 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001001110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -82 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001001111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -81 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001010000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -80 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001010001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -79 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001010010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -78 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001010011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -77 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001010100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -76 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001010101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -75 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001010110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -74 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001010111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -73 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001011000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -72 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001011001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -71 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001011010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -70 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001011011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -69 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001011100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -68 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001011101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -67 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001011110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -66 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001011111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -65 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001100000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -64 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001100001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -63 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001100010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -62 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001100011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -61 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001100100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -60 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001100101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -59 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001100110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -58 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001100111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -57 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001101000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -56 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001101001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -55 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001101010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -54 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001101011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -53 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001101100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -52 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001101101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -51 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001101110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -50 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001101111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col -49 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "001110000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -48 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001110001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -47 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001110010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -46 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001110011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -45 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001110100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -44 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001110101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -43 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001110110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -42 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001110111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -41 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -40 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -39 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -38 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -37 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -36 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -35 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -34 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "001111111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -33 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -32 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -31 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -30 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -29 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -28 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -27 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -26 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010000111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -25 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -24 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -23 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -22 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -21 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -20 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -19 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -18 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010001111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -17 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -16 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -15 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -14 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -13 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -12 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -11 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -10 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010010111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -9 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -8 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -7 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -6 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -5 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -4 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -3 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -2 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010011111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col -1 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 0 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 1 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 2 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 3 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 4 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 5 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 6 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010100111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 7 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 8 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 9 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 10 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 11 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 12 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 13 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 14 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010101111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 15 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 16 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 17 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 18 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 19 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 20 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 21 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 22 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010110111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 23 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 24 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 25 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 26 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 27 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 28 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 29 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 30 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "010111111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 31 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 32 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 33 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 34 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 35 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 36 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 37 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 38 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011000111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 39 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011001000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 40 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011001001"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 41 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011001010"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 42 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011001011"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 43 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011001100"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 44 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011001101"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 45 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011001110"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 46 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011001111"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 47 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011010000"; wait for 40 ns;
		assert screen_col_color = "000000011111001111" report "row -40, col 48 r,g,b = 0, 31, 15" severity error;
		screen_col_to_read <= "011010001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 49 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011010010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 50 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011010011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 51 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011010100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 52 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011010101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 53 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011010110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 54 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011010111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 55 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011011000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 56 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011011001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 57 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011011010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 58 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011011011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 59 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011011100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 60 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011011101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 61 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011011110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 62 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011011111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 63 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011100000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 64 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011100001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 65 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011100010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 66 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011100011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 67 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011100100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 68 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011100101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 69 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011100110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 70 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011100111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 71 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011101000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 72 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011101001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 73 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011101010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 74 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011101011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 75 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011101100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 76 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011101101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 77 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011101110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 78 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011101111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 79 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011110000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 80 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011110001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 81 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011110010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 82 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011110011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 83 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011110100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 84 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011110101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 85 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011110110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 86 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011110111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 87 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011111000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 88 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011111001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 89 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011111010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 90 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011111011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 91 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011111100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 92 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011111101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 93 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011111110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 94 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "011111111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 95 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100000000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 96 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100000001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 97 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100000010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 98 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100000011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 99 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100000100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 100 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100000101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 101 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100000110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 102 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100000111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 103 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100001000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 104 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100001001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 105 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100001010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 106 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100001011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 107 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100001100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 108 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100001101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 109 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100001110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 110 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100001111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 111 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100010000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 112 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100010001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 113 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100010010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 114 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100010011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 115 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100010100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 116 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100010101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 117 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100010110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 118 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100010111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 119 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100011000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 120 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100011001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 121 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100011010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 122 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100011011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 123 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100011100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 124 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100011101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 125 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100011110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 126 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100011111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 127 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100100000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 128 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100100001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 129 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100100010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 130 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100100011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 131 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100100100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 132 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100100101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 133 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100100110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 134 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100100111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 135 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100101000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 136 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100101001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 137 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100101010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 138 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100101011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 139 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100101100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 140 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100101101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 141 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100101110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 142 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100101111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 143 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100110000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 144 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100110001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 145 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100110010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 146 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100110011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 147 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100110100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 148 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100110101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 149 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100110110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 150 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100110111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 151 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100111000"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 152 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100111001"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 153 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100111010"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 154 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100111011"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 155 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100111100"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 156 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100111101"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 157 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100111110"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 158 r,g,b = 31, 31, 15" severity error;
		screen_col_to_read <= "100111111"; wait for 40 ns;
		assert screen_col_color = "011111011111001111" report "row -40, col 159 r,g,b = 31, 31, 15" severity error;
		wait until done = '1'; row <= "111010110";
		wait until done = '1'; row <= "111010101";
		wait until done = '1'; row <= "111010100";
		wait until done = '1'; row <= "111010011";
		wait until done = '1'; row <= "111010010";
		wait until done = '1'; row <= "111010001";
		wait until done = '1'; row <= "111010000";
		wait until done = '1'; row <= "111001111";
		wait until done = '1'; row <= "111001110";
		wait until done = '1';
		
		report "Don't worry, it's just the end of the simulation." severity failure;
	end process;
end test;
